import logging
from typing import List, Set, Collection
from is_api import entities

from portal.database import Course, User, queries, Group
from portal.service.general import GeneralService
from portal.tools.is_api_adapter import IsApiFactory, IsApiWrapper

log = logging.getLogger(__name__)


def _users_to_ucos_set(users: List[User]) -> Set[int]:
    return set(user.uco for user in users)


class IsApiService(GeneralService):
    @property
    def roles(self):
        from portal.service.roles import RoleService
        return RoleService

    @property
    def groups(self):
        from portal.service.groups import GroupService
        return GroupService

    @property
    def is_factory(self) -> IsApiFactory:
        from portal import is_api_factory
        return is_api_factory

    @property
    def course(self) -> Course:
        return self._entity

    @property
    def is_api(self) -> IsApiWrapper:
        return self.is_factory.get_instance(self.course)

    def import_students(self, role_name: str = 'student'):
        """Import students to course from IS MU
        Args:
            role_name: Name of the role for the students

        Returns:

        """
        students = self.is_api.course_students
        ucos = [student.uco for student in students.students]
        self._import_users(ucos, role_name)

    def import_teachers(self, role_name: str = 'teacher'):
        """Import teachers to course from IS MU
        Args:
            role_name: Name of the role for teachers

        Returns:

        """
        teachers = self.is_api.course_teachers
        ucos = [teacher.uco for teacher in teachers]
        self._import_users(ucos, role_name)

    def sync_seminaries(self):
        """Synchronisation for the seminaries: for given course
        """
        seminaries = self.is_api.course_seminars_list
        for seminary in seminaries:
            self._sync_seminary(seminary)

    def _import_users(self, uco_list: List[int], role_name: str):
        log.info(f"Importing users to course {self.course.log_name} "
                 f"with role {role_name}: {uco_list}")
        role = self.find.role(course=self.course, identifier=role_name)
        for uco in uco_list:
            user = queries.user_by_uco(uco).first()
            if not user:
                log.info(f"User with uco: {uco} does not exists - SKIPPING.")
                return
            self.roles(role).add_client(user)

    def _sync_seminary(self, seminary: entities.Seminar):
        group = self._get_or_create_group(seminary)
        s_uco = self._extract_ucos_students(seminary)
        t_uco = self._extract_ucos_teachers(seminary)
        current_is = set(*s_uco, *t_uco)
        current_portal = _users_to_ucos_set(group.users)
        remove = current_portal - current_is
        add = current_is - current_portal
        log.info(f"[SYNC] Seminary SYNC - remove: {remove}, add: {add}")
        self._mod_person_in_group(group, add, 'add_user')
        self._mod_person_in_group(group, remove, 'remove_user')

    def _get_or_create_group(self, seminary: entities.Seminar):
        group = self.find.group(self.course, seminary.label)
        if not group:
            params = dict(codename=seminary.label, name=f"Seminary {seminary.label}")
            group = self.groups.create(course=self.course, **params)
        return group

    def _extract_ucos_students(self, seminary: entities.Seminar):
        items = self.is_api.is_client.seminar_list_students(seminars=[seminary.label])
        return [item.uco for item in items.seminars[0].students]

    def _mod_person_in_group(self, group: Group, collection: Collection[int], action: str):
        for uco in collection:
            user: User = queries.user_by_uco(uco).first()
            if not user:
                log.info(f"User with uco: {uco} does not exists - SKIPPING.")
                return
            log.info(f"User with uco {uco} ({user.log_name}), "
                     f"action - {action} on group: {group.log_name}")
            group_service = self.groups(group)
            action_method = getattr(group_service, action)
            action_method(user)

    def _extract_ucos_teachers(self, seminary: entities.Seminar):
        items = self.is_api.is_client.seminar_list_teachers(seminars=[seminary.label])
        return [item.uco for item in items.seminars[0].teachers]
